#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>

#ifdef __USAGE
%C - manipulate VMware

%C  [options]
Options
-V         get VMware version (default)
-t         get host time
  -f format  strftime format
-s         get host screen size
-g         get host clipboard
-p         put from stdin to host clipboard
-e         enumerate devices
#endif

/*
 * The following info from,
 * 
 * chitchat.at.infoseek.co.jp/vmware/vmtools.html
 * Author: Ken Kato
 * chitchat-lj@infoseek.jp
 * 
 * You may freely use, modify, redistribute or trash them in whatever way
 * you like. Success or bug reports, suggestions, encouragement and
 * compliments are greatly appreciated though. Complaints, accusation and
 * abuse are not.
 */
#define VMWARE_MAGIC        0x564D5868      /* Backdoor magic number        */
#define VMWARE_PORT         0x5658          /* Backdoor port number         */

#define VMCMD_APM_FUNCTION  0x02            /* APM function?                */
#define VMCMD_GET_CURSOR    0x04            /* Get cursor position          */
#define VMCMD_SET_CURSOR    0x05            /* Set cursor position          */
#define VMCMD_GET_CLIPLEN   0x06            /* host->guest copy length      */
#define VMCMD_GET_CLIPDATA  0x07            /* host->guest copy data        */
#define VMCMD_SET_CLIPLEN   0x08            /* guest->host copy length      */
#define VMCMD_SET_CLIPDATA  0x09            /* guest->host copy date        */
#define VMCMD_GET_VERSION   0x0a            /* Get version number           */
#define VMCMD_GET_DEVICE    0x0b            /* Get device information       */
#define VMCMD_SET_DEVICE    0x0c            /* Set device state             */
#define VMCMD_GET_SETTING   0x0d            /* Get preference settings      */
#define VMCMD_SET_SETTING   0x0e            /* Set Preference settings      */
#define VMCMD_GET_HOSTRES   0x0f            /* Get host's screen size       */
#define VMCMD_POPUP_DIALOG  0x12            /* OS not found message popup   */
#define VMCMD_GET_VMCONFIG  0x14            /* Get VM config information    */
#define VMCMD_GET_TIME      0x17            /* Get host's time (GMT)        */
#define VMCMD_TRANSFER      0x1e            /* Generic data transfer        */

/*
 *  Preference bitmask (for VMCMD_GET_SETTING, VMCMD_SET_SETTING)
 */
#define VMPREF_GRAB_MOVE    0x0001          /* grab by cursor movement      */
#define VMPREF_UNGRAB_MOVE  0x0002          /* ungrab by cursor movement    */
#define VMPREF_SCROLL_MOVE  0x0004          /* scroll by cursor movement    */
#define VMPREF_COPY_PASTE   0x0010          /* copy and paste               */
#define VMPREF_FULLSCREEN   0x0040          /* full screen (read only)      */
#define VMPREF_ENTER_FULL   0x0080          /* full screen (write only)     */
#define VMPREF_SYNCRONIZE   0x0400          /* time syncronization          */

#define VMCLIP_MAX_LENGTH   65355           /* maximum copy & paste length  */

#ifndef __QNXNTO__
extern void vm_cmd(unsigned *p);
#pragma aux vm_cmd = \
	"mov eax, dword ptr [edi]" \
	"mov ebx, dword ptr [edi+4]" \
	"mov ecx, dword ptr [edi+8]" \
	"mov edx, dword ptr [edi+12]" \
	"in eax, dx" \
	"mov dword ptr [edi],    eax" \
	"mov dword ptr [edi+4],  ebx" \
	"mov dword ptr [edi+8],  ecx" \
	"mov dword ptr [edi+12], edx" \
	parm [edi] \
	modify [eax ebx ecx edx];
#else
extern void vm_cmd(unsigned *p) {
        __asm__  ("in %%dx, %%eax"
        : "=a"(p[0]), "=b"(p[1]), "=c"(p[2]), "=d"(p[3])
        : "a"(p[0]), "b"(p[1]), "c"(p[2]), "d"(p[3])
        );
}
#endif

void 
segfault()
{
	printf("Not running inside VMware.\n");
	exit(1);
}

int
main(int argc, char *argv[])
{
	int             i;
	int             fun = VMCMD_GET_VERSION;
	unsigned int    regs[4];
	unsigned char  *format = NULL;

	signal(SIGSEGV, segfault);
	while ((i = getopt(argc, argv, "Vtf:sgpe")) != -1) {
		switch (i) {
		case 't':
			fun = VMCMD_GET_TIME;
			break;
		case 'f':
			format = optarg;
			break;
		case 's':
			fun = VMCMD_GET_HOSTRES;
			break;
		case 'g':
			fun = VMCMD_GET_CLIPDATA;
			break;
		case 'p':
			fun = VMCMD_SET_CLIPDATA;
			break;
		case 'e':
			fun = VMCMD_GET_DEVICE;
			break;
		case 'V':
		default:
			fun = VMCMD_GET_VERSION;
			break;
		}
	}
	switch (fun) {
	case VMCMD_GET_TIME:
		regs[0] = VMWARE_MAGIC;
		regs[1] = 0xc0de;	/* don't care */
		regs[2] = fun;
		regs[3] = VMWARE_PORT;
		vm_cmd(regs);
		if (format) {
			char            buffer[80];

			strftime(buffer, 80, format, localtime(&regs[0]));
			printf("%s\n", buffer);
		} else {
			printf("%d\n",regs[0]);
		}
		break;
	case VMCMD_GET_HOSTRES:
		regs[0] = VMWARE_MAGIC;
		regs[1] = 0xc0de;	/* don't care */
		regs[2] = fun;
		regs[3] = VMWARE_PORT;
		vm_cmd(regs);
		printf("%dx%d\n", regs[0] >> 16, regs[0] & 0xffff);
		break;
	case VMCMD_GET_CLIPDATA:
		{
			int             length;
			unsigned char   *text;

			regs[0] = VMWARE_MAGIC;
			regs[1] = 0xc0de;	/* don't care */
			regs[2] = VMCMD_GET_SETTING;
			regs[3] = VMWARE_PORT;
			vm_cmd(regs);
			if (!(regs[0] & VMPREF_COPY_PASTE)) {
				printf("clipboard is disabled.\n");
				return 1;
			}
			length = 0;
			regs[0] = VMWARE_MAGIC;
			regs[1] = 0xc0de;	/* don't care */
			regs[2] = VMCMD_GET_CLIPLEN;
			regs[3] = VMWARE_PORT;
			vm_cmd(regs);
			length = regs[0];
			if (length) {
				for (;;) {
					regs[0] = VMWARE_MAGIC;
					regs[1] = 0xc0de;	/* don't care */
					regs[2] = VMCMD_GET_CLIPDATA;
					regs[3] = VMWARE_PORT;
					vm_cmd(regs);
					text = (unsigned char *)&regs[0];
					if (length < 4) {
						text[length] = '\0';
					}
					for (i = 0; i < 4; i++) {
						if (text[i]) {
							putchar(text[i]);
						} else {
							length = 0;
							break;
						}
					}
					if (length <= 4) {
						break;
					}
					length -= 4;
				}
			}
		}
		break;
	case VMCMD_SET_CLIPDATA:
		{
			char            buf[VMCLIP_MAX_LENGTH + 1];
			int             length = 0, maxlen = VMCLIP_MAX_LENGTH;

			regs[0] = VMWARE_MAGIC;
			regs[1] = 0xc0de;	/* don't care */
			regs[2] = VMCMD_GET_SETTING;
			regs[3] = VMWARE_PORT;
			vm_cmd(regs);
			if (!(regs[0] & VMPREF_COPY_PASTE)) {
				printf("clipboard is disabled.\n");
				return 1;
			}
			while ((buf[length] = getchar()) != EOF) {
				if (buf[length] == '\n') {
					maxlen--;
				}
				if (++length >= maxlen) {
					break;
				}
			}
			if (length) {
				int             i = 0;
				buf[length++] = '\0';
				regs[0] = VMWARE_MAGIC;
				regs[1] = length;
				regs[2] = VMCMD_SET_CLIPLEN;
				regs[3] = VMWARE_PORT;
				vm_cmd(regs);
				while (i < length) {
					unsigned int   *pi = (unsigned int *) &buf[i];
					regs[0] = VMWARE_MAGIC;
					regs[1] = *pi;
					regs[2] = VMCMD_SET_CLIPDATA;
					regs[3] = VMWARE_PORT;
					vm_cmd(regs);
					i += 4;
				}
			}
		}
		break;
	case VMCMD_GET_DEVICE:
		{
			int i;
			char buf[40];

			for (i = 0; i < 32; i++) {
				int pos = 0;

				memset(buf,0,sizeof(buf));
				while (pos < 40) {
					regs[0] = VMWARE_MAGIC;
					regs[1] = (i<<16) | pos;
					regs[2] = VMCMD_GET_DEVICE;
					regs[3] = VMWARE_PORT;
					vm_cmd(regs);
					*((unsigned int *)(buf + pos)) = regs[1];
					pos += 4;
				}
				if (buf[0] == 0) {
					continue;
				}
				printf("  %c%d\t%s\n", buf[36] ? '+' : ' ', i, buf);
			}
		}
		break;
	case VMCMD_GET_VERSION:
	default:
		regs[0] = VMWARE_MAGIC;
		regs[1] = 0xc0de;	/* don't care */
		regs[2] = fun;
		regs[3] = VMWARE_PORT;
		vm_cmd(regs);
		printf("%d,%d\n", regs[0], regs[2]);
		break;
	}
	return 0;
}
