#include <stdio.h>
#include <unistd.h>
#include <signal.h>
#include <time.h>
#include <Ph.h>
#include <Pt.h>

#ifdef __USAGE
%C - Daemon to sync clipboards and time with host system

#endif

/*
 * Compile in QNX6 with 'qcc -o vmtoold vmtoold.c -lph'
 * Run as root if you want your clock updated
 * 
 * Compile in QNX4 with 'cc -o vmtoold -T1 vmtoold.c -N1M -lphoton'
 *
 * The following info from,
 * 
 * chitchat.at.infoseek.co.jp/vmware/vmtools.html
 * Author: Ken Kato
 * chitchat-lj@infoseek.jp
 * 
 * You may freely use, modify, redistribute or trash them in whatever way
 * you like. Success or bug reports, suggestions, encouragement and
 * compliments are greatly appreciated though. Complaints, accusation and
 * abuse are not.
 */
#define VMWARE_MAGIC        0x564D5868      /* Backdoor magic number        */
#define VMWARE_PORT         0x5658          /* Backdoor port number         */

#define VMCMD_APM_FUNCTION  0x02            /* APM function?                */
#define VMCMD_GET_CURSOR    0x04            /* Get cursor position          */
#define VMCMD_SET_CURSOR    0x05            /* Set cursor position          */
#define VMCMD_GET_CLIPLEN   0x06            /* host->guest copy length      */
#define VMCMD_GET_CLIPDATA  0x07            /* host->guest copy data        */
#define VMCMD_SET_CLIPLEN   0x08            /* guest->host copy length      */
#define VMCMD_SET_CLIPDATA  0x09            /* guest->host copy date        */
#define VMCMD_GET_VERSION   0x0a            /* Get version number           */
#define VMCMD_GET_DEVICE    0x0b            /* Get device information       */
#define VMCMD_SET_DEVICE    0x0c            /* Set device state             */
#define VMCMD_GET_SETTING   0x0d            /* Get preference settings      */
#define VMCMD_SET_SETTING   0x0e            /* Set Preference settings      */
#define VMCMD_GET_HOSTRES   0x0f            /* Get host's screen size       */
#define VMCMD_POPUP_DIALOG  0x12            /* OS not found message popup   */
#define VMCMD_GET_VMCONFIG  0x14            /* Get VM config information    */
#define VMCMD_GET_TIME      0x17            /* Get host's time (GMT)        */
#define VMCMD_TRANSFER      0x1e            /* Generic data transfer        */

/*
 *  Preference bitmask (for VMCMD_GET_SETTING, VMCMD_SET_SETTING)
 */
#define VMPREF_GRAB_MOVE    0x0001          /* grab by cursor movement      */
#define VMPREF_UNGRAB_MOVE  0x0002          /* ungrab by cursor movement    */
#define VMPREF_SCROLL_MOVE  0x0004          /* scroll by cursor movement    */
#define VMPREF_COPY_PASTE   0x0010          /* copy and paste               */
#define VMPREF_FULLSCREEN   0x0040          /* full screen (read only)      */
#define VMPREF_ENTER_FULL   0x0080          /* full screen (write only)     */
#define VMPREF_SYNCRONIZE   0x0400          /* time syncronization          */

#define VMCLIP_MAX_LENGTH   65355           /* maximum copy & paste length  */

#ifndef __QNXNTO__
extern void vm_cmd(unsigned *p);
#pragma aux vm_cmd = \
	"mov eax, dword ptr [edi]" \
	"mov ebx, dword ptr [edi+4]" \
	"mov ecx, dword ptr [edi+8]" \
	"mov edx, dword ptr [edi+12]" \
	"in eax, dx" \
	"mov dword ptr [edi],    eax" \
	"mov dword ptr [edi+4],  ebx" \
	"mov dword ptr [edi+8],  ecx" \
	"mov dword ptr [edi+12], edx" \
	parm [edi] \
	modify [eax ebx ecx edx];
#define PhInputGroup(x) (1)
#define PtExit(x) exit(x)
#else
extern void vm_cmd(unsigned *p) {
        __asm__  ("in %%dx, %%eax"
        : "=a"(p[0]), "=b"(p[1]), "=c"(p[2]), "=d"(p[3])
        : "a"(p[0]), "b"(p[1]), "c"(p[2]), "d"(p[3])
        );
}
#endif

void 
segfault()
{
	printf("Not running inside VMware.\n");
	exit(1);
}

char *host, *current, *photon;

void update_time();
char *get_host_clip();
void set_host_clip(char *clip);

int
main(int argc, char *argv[])
{
  int             i = 0;

  signal(SIGSEGV, segfault);

  if(PtInit(NULL) == -1)
    PtExit(-1);

  current = PhClipboardPasteString( PhInputGroup(NULL));

  while(1)
  {
    i++;
    sleep(3);
    if( !(i % 20 ))
    update_time();
    
    photon = PhClipboardPasteString( PhInputGroup(NULL));
    host = get_host_clip();

    if( !current && photon )
      {
        // current version is null, but photon is valid
        current = photon; // should only happen once
      }

     if( current && photon )
      {
        /*        printf( "C%d: %s\nP%d: %s\nH%d: %s\n", 
                current->length, current->data, 
                photon->length, photon->data, 
                host.length, host.data);
        */
        if(strcmp(current, photon) != 0)
          {
            // new photon clip
            free(current);
            current = photon;
            set_host_clip(current);
          }
        else if(host && strcmp(current, host) != 0)
          {
            free(current);
			current = host;
            PhClipboardCopyString(PhInputGroup(NULL), current);
          }
        else
          {
            if(photon)
              free(photon);
            if(host)
              free(host);
          }
      }
 }
}

/* Get host time and update QNX time */
void update_time()
{
  unsigned int    regs[4];
  struct timespec stime;
 
  regs[0] = VMWARE_MAGIC;
  regs[1] = 0xc0de;	/* don't care */
  regs[2] = VMCMD_GET_TIME;
  regs[3] = VMWARE_PORT;
  vm_cmd(regs);
   stime.tv_sec = regs[0];
  stime.tv_nsec = 0;
  clock_settime(CLOCK_REALTIME, &stime);
}

/* Get host clipboard */
char * get_host_clip()
{
  int             i;
  unsigned int    regs[4];
  int             length;
  unsigned char  *text, *ptr;
  unsigned char  *clip;

  regs[0] = VMWARE_MAGIC;
  regs[1] = 0xc0de;	/* don't care */
  regs[2] = VMCMD_GET_SETTING;
  regs[3] = VMWARE_PORT;
  vm_cmd(regs);
  if (!(regs[0] & VMPREF_COPY_PASTE)) 
    {
      return NULL;
    }
  length = 0;
  regs[0] = VMWARE_MAGIC;
  regs[1] = 0xc0de;	/* don't care */
  regs[2] = VMCMD_GET_CLIPLEN;
  regs[3] = VMWARE_PORT;
  vm_cmd(regs);
  length = regs[0];
  if (length) 
    {
      clip = calloc(1, length+1);
      if(clip == NULL)
        {
          return NULL;
        }
      ptr = clip;
      for (;;) 
        {
          regs[0] = VMWARE_MAGIC;
          regs[1] = 0xc0de;	/* don't care */
          regs[2] = VMCMD_GET_CLIPDATA;
          regs[3] = VMWARE_PORT;
          vm_cmd(regs);
          text = (unsigned char *)&regs[0];
          if (length < 4) 
            {
              text[length] = '\0';
            }
          for (i = 0; i < 4; i++) 
            {
              if (text[i]) 
                {
                  *ptr = text[i];
                  ptr++;
                } 
              else 
                {
                  length = 0;
                  break;
                }
            }
          if (length <= 4) 
            {
              break;
            }
          length -= 4;
        }
      return clip;
    }
  else
    {
      return NULL;
    }
}

/* Set host clipboard */
void set_host_clip(char *clip)
{
  unsigned int    regs[4];
  int             maxlen = VMCLIP_MAX_LENGTH;
  int             i, length;
  char           *text = clip;

  regs[0] = VMWARE_MAGIC;
  regs[1] = 0xc0de;	/* don't care */
  regs[2] = VMCMD_GET_SETTING;
  regs[3] = VMWARE_PORT;
  vm_cmd(regs);
  if (!(regs[0] & VMPREF_COPY_PASTE)) 
    {
      return ;
    }

  length = strlen(clip);
  if(length > maxlen)
    {
      length = maxlen;
    }

  if (length) 
    {
      int             i = 0;
      regs[0] = VMWARE_MAGIC;
      regs[1] = length;
      regs[2] = VMCMD_SET_CLIPLEN;
      regs[3] = VMWARE_PORT;
      vm_cmd(regs);
      while (i < length) 
        {
          unsigned int   *pi = (unsigned int *) &text[i];
          regs[0] = VMWARE_MAGIC;
          regs[1] = *pi;
          regs[2] = VMCMD_SET_CLIPDATA;
          regs[3] = VMWARE_PORT;
          vm_cmd(regs);
          i += 4;
        }
    }
}
